package cn.com.poc.expose.websocket.handler;

import cn.com.poc.common.constant.CommonConstant;
import cn.com.poc.common.entity.media.MediaType;
import cn.com.poc.common.service.BosConfigService;
import cn.com.poc.common.utils.*;
import cn.com.poc.communication.entity.*;
import cn.com.poc.communication.service.*;
import cn.com.poc.expose.dto.DialoguesWsResponse;
import cn.com.poc.expose.dto.WSCommunicationResult;
import cn.com.poc.expose.websocket.exception.WebsocketException;
import cn.com.poc.message.entity.CommunicationRecordMessage;
import cn.com.poc.message.service.CommunicationRecordProducerService;
import cn.com.poc.thirdparty.common.constant.MessageRoleConstant;
import cn.com.poc.thirdparty.resource.demand.ai.common.domain.Message;
import cn.com.poc.thirdparty.resource.demand.ai.common.domain.MultiContent;
import cn.com.poc.thirdparty.resource.demand.ai.entity.largemodel.LargeModelDemandResult;
import cn.com.poc.thirdparty.resource.demand.ai.entity.largemodel.LargeModelResponse;
import cn.com.poc.thirdparty.service.LLMService;
import cn.com.yict.framemax.core.config.Config;
import cn.com.yict.framemax.core.exception.BusinessException;
import cn.com.yict.framemax.frame.service.FmxParamConfigService;
import com.alibaba.fastjson.JSONObject;
import com.github.houbb.opencc4j.util.ZhConverterUtil;
import com.tencent.SpeechClient;
import com.tencent.tts.model.SpeechSynthesisRequest;
import com.tencent.tts.model.SpeechSynthesisResponse;
import com.tencent.tts.service.SpeechSynthesisListener;
import com.tencent.tts.service.SpeechSynthesizer;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.java_websocket.WebSocket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;


public class CantoneseCommunicationHandler extends AbstractWsHandler {

    final private String EVENT_STREAM_PREFIX = "data: ";

    final private Logger logger = LoggerFactory.getLogger(CantoneseCommunicationHandler.class);

    final private LLMService llmService;

    final private BizCommunicationThemeService bizCommunicationThemeService;

    final private BizCommunicationService bizCommunicationService;

    final private BizCommunicationRecordService bizCommunicationRecordService;

    final private BizAiPromptService bizAiPromptService;

    final private BizSpeechToTextService bizSpeechToTextService;

    final private CommunicationRecordProducerService communicationRecordProducerService;

    final private BosConfigService bosConfigService;

    final private FmxParamConfigService fmxParamConfigService;

    final private StringBuilder aiRestBuffer = new StringBuilder();

    private SpeechSynthesizer speechSynthesizer;
    private Integer VOLUME = 0;
    private Float SPEED = 0F;


    public CantoneseCommunicationHandler() {
        llmService = SpringUtils.getBean(LLMService.class);
        bizCommunicationThemeService = SpringUtils.getBean(BizCommunicationThemeService.class);
        bizCommunicationService = SpringUtils.getBean(BizCommunicationService.class);
        bizCommunicationRecordService = SpringUtils.getBean(BizCommunicationRecordService.class);
        bizAiPromptService = SpringUtils.getBean(BizAiPromptService.class);
        bizSpeechToTextService = SpringUtils.getBean(BizSpeechToTextService.class);
        communicationRecordProducerService = SpringUtils.getBean(CommunicationRecordProducerService.class);
        bosConfigService = SpringUtils.getBean(BosConfigService.class);
        fmxParamConfigService = SpringUtils.getBean(FmxParamConfigService.class);
    }


    @Override
    public void doHandler(WebSocket webSocket, String message) {
        Assert.notBlank(message, "消息不能为空");
        logger.info("收到客户端消息:{}", message);
        DialoguesWsResponse dialoguesWsResponse = JSONObject.parseObject(message, DialoguesWsResponse.class);
        Assert.notNull(dialoguesWsResponse.getCommunicationId(), "Communication ID不能为空");
        Assert.notNull(dialoguesWsResponse.getVoiceId(), "Voice ID不能为空");
        String communicationId = dialoguesWsResponse.getCommunicationId();
        try {
            String userContent = dialoguesWsResponse.getUserContent();
            String userSpeechUrl = "";

            if (StringUtils.isNoneBlank(dialoguesWsResponse.getTaskId())) {
                BizSpeechToTextEntity bizSpeechToTextEntity = bizSpeechToTextService.getByTaskId(dialoguesWsResponse.getTaskId());
                if (null == bizSpeechToTextEntity) {
                    throw new BusinessException("Speech To Text Task ID is not found");
                }
                userContent = bizSpeechToTextEntity.getContent();
                userSpeechUrl = bizSpeechToTextEntity.getUrl();
            }

            if (StringUtils.isBlank(userContent)) {
                throw new BusinessException("Content is empty");
            }

            BizCommunicationEntity bizCommunicationEntity = bizCommunicationService.getByCommunicationId(communicationId);
            if (null == bizCommunicationEntity) {
                throw new BusinessException("Communication ID is not found");
            }

            // 构造大模型请求参数
            LargeModelResponse largeModelResponse = chatResponseBuild(bizCommunicationEntity.getThemeCode(), userContent, communicationId);
            //初始化TTS
            initTTS(communicationId, webSocket, VOLUME, SPEED);
            //保存[用户]对话记录
            recordMessage(userContent, communicationId, System.currentTimeMillis(), MessageRoleConstant.user, userSpeechUrl);
            //调用大模型
            invokeLLMAndWrite(largeModelResponse, webSocket);
        } catch (Exception e) {
            webSocket.send(e.getMessage());
            webSocket.close();
            throw new WebsocketException(e.getMessage());
        }
    }


    /**
     * 构建大模型请求实体
     *
     * @param themeCode   主题代码
     * @param userContent 用户内容
     * @return 大模型调用实体
     * @throws Exception
     */
    private LargeModelResponse chatResponseBuild(String themeCode, String userContent, String communicationId) throws Exception {
        logger.info("---------- chat response build start,themeCode:{},userContent:{},timestamp:{} ---------", themeCode, userContent, System.currentTimeMillis());
        if (StringUtils.isBlank(userContent)) {
            throw new WebsocketException("user content is empty");
        }
        List<Message> messages = new ArrayList<>();
        LargeModelResponse request = new LargeModelResponse();

        //获取主题配置
        BizCommunicationThemeEntity themeEntity = bizCommunicationThemeService.getByThemeCode(themeCode);
        if (themeEntity != null) {
            String promptRelation = themeEntity.getPromptRelation();
            //获取主题提示语
            BizAiPromptEntity promptEntity = bizAiPromptService.getByPromptCode(promptRelation);
            if (StringUtils.isNotBlank(promptEntity.getPromptValue())) {
                Message systemMessage = new Message();
                systemMessage.setRole(MessageRoleConstant.system);
                systemMessage.setContent(getContent(promptEntity.getPromptValue()));
                messages.add(systemMessage);
            }
        }

        // 历史记录（上下文）
        List<BizCommunicationRecordEntity> bizCommunicationRecordEntities = bizCommunicationRecordService.get(communicationId);
        if (CollectionUtils.isNotEmpty(bizCommunicationRecordEntities)) {
            for (BizCommunicationRecordEntity bizCommunicationRecordEntity : bizCommunicationRecordEntities) {
                String role = bizCommunicationRecordEntity.getRole();
                String content = bizCommunicationRecordEntity.getContent();
                Message message = new Message();
                message.setRole(role);
                message.setContent(getContent(content));
                messages.add(message);
            }
            // 上下文校验清洗
            BizCommunicationRecordEntity bizCommunicationRecordEntity = bizCommunicationRecordEntities.get(bizCommunicationRecordEntities.size() - 1);
            //删除无用数据
            if (bizCommunicationRecordEntity.getRole().equals(MessageRoleConstant.user)) {
                messages.remove(messages.size() - 1);
                bizCommunicationRecordService.del(bizCommunicationRecordEntity.getId());
            }
        }


        //构建大模型实体
        Message userMessage = new Message();
        userMessage.setRole(MessageRoleConstant.user);
        userMessage.setContent(getContent(userContent));
        messages.add(userMessage);

        Message[] array = new Message[messages.size()];
        messages.toArray(array);

        request.setMessages(array);
        request.setStream(true);
        request.setModel(fmxParamConfigService.getParam("cantonese.communication.model"));
        return request;
    }

    private static List<MultiContent> getContent(String content) {
        MultiContent multiContent = new MultiContent();
        multiContent.setType("text");
        multiContent.setText(content);
        List<MultiContent> multiContents = new ArrayList<>();
        multiContents.add(multiContent);
        return multiContents;
    }


    /**
     * 调用大模型
     *
     * @param request
     * @return
     * @throws Exception
     */
    private String invokeLLMAndWrite(LargeModelResponse request, WebSocket webSocket) throws Exception {
        //调用GPT
        BufferedReader bufferedReader = llmService.chatChunk(request);

        //判断大模型是否调用成功
        if (bufferedReader == null) {
            throw new WebsocketException("llm is error");
        }

        //大模型数据响应读取
        String res;
        while ((res = bufferedReader.readLine()) != null) {
            if (StringUtils.isBlank(res)) {
                continue;
            }
            logger.debug("------- communication llm content:{} ---------", res);
            LargeModelDemandResult aiResult = JsonUtils.deSerialize(res.replaceFirst(EVENT_STREAM_PREFIX, StringUtils.EMPTY), LargeModelDemandResult.class);
            if (aiResult == null) {
                continue;
            }

            if (!aiResult.getCode().equals("0")) {
                throw new WebsocketException("llm is error");
            }

            if (aiResult.getMessage() != null && StringUtils.isNotBlank(aiResult.getMessage())) {
                String content = aiResult.getMessage();
                aiRestBuffer.append(content);
                content = content.replaceAll("\"", "\\\\\"").replaceAll("\n", StringUtils.SPACE);
                WSCommunicationResult result = new WSCommunicationResult();
                result.setReplyContent(content);
                result.setTransContent(ZhConverterUtil.convertToTraditional(content));
                webSocket.send(JSONObject.toJSONString(result));
            }
        }
        //调用TTS
        textToSpeech(aiRestBuffer.toString());
        logger.debug("------------ communication llm end , content : {} --------------", aiRestBuffer);
        return aiRestBuffer.toString();
    }

    /**
     * 初始化TTS
     */
    private void initTTS(String communicationId, WebSocket webSocket, Integer volume, Float speed) throws IOException {
        //从配置文件读取密钥
        String appId = Config.get("tencent.speech.synthesizer.appid");
        String secretId = Config.get("tencent.speech.synthesizer.secretId");
        String secretKey = Config.get("tencent.speech.synthesizer.secretKey");
        //创建SpeechSynthesizerClient实例，目前是单例
        SpeechClient client = SpeechClient.newInstance(appId, secretId, secretKey);
        //初始化SpeechSynthesizerRequest，SpeechSynthesizerRequest包含请求参数
        SpeechSynthesisRequest request = SpeechSynthesisRequest.initialize();
        request.setSampleRate(8000);

        request.setCodec(MediaType.voice_mp3.fileSuffix());
        request.setVolume(volume);
        request.setVoiceType(101019);
        //使用客户端client创建语音合成实例
        speechSynthesizer = client.newSpeechSynthesizer(request, new SpeechSynthesisListener() {

            List<byte[]> audioBytes = new ArrayList<>();

            AtomicInteger sessionId = new AtomicInteger(0);
            File tempAudioFile = File.createTempFile(UUIDTool.getUUID(), MediaType.voice_mp3.fileSuffix());

            @Override
            public void onComplete(SpeechSynthesisResponse response) {
                logger.info("onComplete");
                try (FileOutputStream fileOutputStream = new FileOutputStream(tempAudioFile, true)) {
                    for (byte[] audioByte : audioBytes) {
                        fileOutputStream.write(audioByte);
                    }
                } catch (IOException e) {
                    logger.error("onComplete:{}", e.getMessage());
                }
                //上传音频
                try (InputStream fileInputStream = Files.newInputStream(tempAudioFile.toPath());) {
                    String uploadUrl = bosConfigService.upload(fileInputStream, MediaType.voice_mp3.fileSuffix(), null);
                    recordMessage(aiRestBuffer.toString(), communicationId, System.currentTimeMillis(), MessageRoleConstant.assistant, uploadUrl);
                    webSocket.send("{\"replyVoiceUrl\":\"" + uploadUrl + "\"}");
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
                webSocket.send("{\"final\":true}");
                webSocket.close();
            }

            @Override
            public void onMessage(byte[] data) {
                sessionId.incrementAndGet();
//                String base64 = Base64.getEncoder().encodeToString(data);
                webSocket.send("{\"audio\":" + Arrays.toString(data) + "}");
                audioBytes.add(data);
            }

            @Override
            public void onFail(SpeechSynthesisResponse response) {
                logger.warn("onFail:{}", response.getMessage());
                webSocket.close();
            }
        });
    }

    /**
     * 调用TTS
     */
    private void textToSpeech(String text) {
        this.speechSynthesizer.synthesisLongText(text);
    }


    /**
     * 日志记录
     *
     * @param userContent         用户内容
     * @param communicationId     会话ID
     * @param userRecordTimestamp 用户回答时间
     * @param role                角色
     * @param speechUrl           音频地址
     */
    private void recordMessage(String userContent, String communicationId, long userRecordTimestamp, String role, String speechUrl) throws Exception {
        BizCommunicationRecordEntity bizCommunicationRecordEntity = new BizCommunicationRecordEntity();
        bizCommunicationRecordEntity.setCommunicationId(communicationId);
        bizCommunicationRecordEntity.setTimestamp(userRecordTimestamp);
        bizCommunicationRecordEntity.setSpeechUrl(speechUrl);
        bizCommunicationRecordEntity.setContent(userContent);
        bizCommunicationRecordEntity.setRole(role);
        bizCommunicationRecordEntity.setAccuracyScore(null);
        bizCommunicationRecordEntity.setFluencyScore(null);
        bizCommunicationRecordEntity.setArticulationScore(null);
        bizCommunicationRecordEntity.setIsDeleted(CommonConstant.IsDeleted.N);
        bizCommunicationRecordEntity.setTransContent(ZhConverterUtil.convertToTraditional(userContent));
        bizCommunicationRecordService.save(bizCommunicationRecordEntity);
    }

}
