package cn.com.poc.thirdparty.resource.tencent.listener;

import cn.com.poc.common.constant.CommonConstant;
import cn.com.poc.common.pool.SpeechToTextThreadPoolExecutor;
import cn.com.poc.common.service.BosConfigService;
import cn.com.poc.common.utils.FileUtils;
import cn.com.poc.common.utils.SpringUtils;
import cn.com.poc.common.utils.UUIDTool;
import cn.com.poc.common.utils.WavMergeUtil;
import cn.com.poc.communication.entity.BizSpeechToTextEntity;
import cn.com.poc.communication.service.BizSpeechToTextService;
import cn.com.poc.expose.websocket.exception.WebsocketException;
import cn.com.yict.framemax.core.exception.BusinessException;
import com.github.houbb.opencc4j.util.ZhConverterUtil;
import com.tencent.asr.model.SpeechRecognitionResponse;
import com.tencent.asr.service.SpeechRecognitionListener;
import com.tencent.core.utils.JsonUtil;
import org.java_websocket.WebSocket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * @author alex.yao
 * @date 2023/12/11
 **/
public class SpeechToTextListener extends SpeechRecognitionListener {

    private final Logger logger = LoggerFactory.getLogger(SpeechToTextListener.class);

    private final Timer timer = new Timer("SPEECH_TO_TEXT", true);

    private final BosConfigService bosConfigService;

    private final BizSpeechToTextService bizSpeechToTextService;

    private WebSocket webSocket;

    private List<byte[]> userAudioByteList;

    private AtomicBoolean isFinal = new AtomicBoolean(false);

    public AtomicBoolean getIsFinal() {
        return isFinal;
    }

    public SpeechToTextListener() {
        //初始化Service服务
        this.bosConfigService = SpringUtils.getBean(BosConfigService.class);
        this.bizSpeechToTextService = SpringUtils.getBean(BizSpeechToTextService.class);
    }

    public void setUserAudioByteList(List<byte[]> userAudioByteList) {
        this.userAudioByteList = userAudioByteList;
    }

    public void setWebSocket(WebSocket webSocket) {
        this.webSocket = webSocket;
    }


    @Override
    public void onSentenceBegin(SpeechRecognitionResponse response) {
//        logger.info("一句话开始:", JsonUtil.toJson(response));
    }

    @Override
    public void onSentenceEnd(SpeechRecognitionResponse response) {
        timer.cancel();
        logger.info("-------------------- onSentenceEnd:{} ----------------------------", JsonUtil.toJson(response));
        if (!isFinal.get()) {
            isFinal.set(true);
            SpeechToTextThreadPoolExecutor.addTask(() -> {
                String taskId = UUIDTool.getUUID();
                String content = response.getResult().getVoiceTextStr();
                this.webSocket.send("{\"userContent\":\"" + content + "\"" +
                        ",\"transContent\":\"" + ZhConverterUtil.convertToTraditional(content) + "\"" +
                        ",\"finish\":true}");
                try {
                    String fileUrl = uploadAudio();

                    webSocket.send("{\"userVoiceUrl\":\"" + fileUrl + "\"}");
                    webSocket.send("{\"taskId\":\"" + taskId + "\"}");
                    webSocket.send("{\"end\":" + true + "}");

                    //保存任务
                    BizSpeechToTextEntity bizSpeechToTextEntity = new BizSpeechToTextEntity();
                    bizSpeechToTextEntity.setTaskId(taskId);
                    bizSpeechToTextEntity.setUrl(fileUrl);
                    bizSpeechToTextEntity.setContent(content);
                    bizSpeechToTextEntity.setTimestamp(System.currentTimeMillis());
                    bizSpeechToTextEntity.setIsDeleted(CommonConstant.IsDeleted.N);
                    bizSpeechToTextService.save(bizSpeechToTextEntity);

                } catch (BusinessException | IOException e) {
                    wsException(e.getMessage(), webSocket);
                } finally {
                    this.webSocket.close();
                }
            });
        }
    }

    @Override
    public void onRecognitionStart(SpeechRecognitionResponse response) {
        logger.info("onRecognitionStart:", JsonUtil.toJson(response));
        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                webSocket.send("{\"reminders\":\"" + true + "\"}");
                timer.cancel();
            }
        }, 1000L * 10);
    }


    @Override
    public void onRecognitionComplete(SpeechRecognitionResponse response) {
        timer.cancel();
        logger.info("onRecognitionComplete:{}", JsonUtil.toJson(response));
    }

    @Override
    public void onRecognitionResultChange(SpeechRecognitionResponse response) {
        timer.cancel();
        logger.info("onRecognitionResultChange:{}", JsonUtil.toJson(response));
        this.webSocket.send("{\"userContent\":\"" + response.getResult().getVoiceTextStr() +
                "\",\"transContent\":\"" + ZhConverterUtil.convertToTraditional(response.getResult().getVoiceTextStr()) + "\"}");
    }


    @Override
    public void onFail(SpeechRecognitionResponse response) {
        timer.cancel();
        logger.warn("onFail:{}", JsonUtil.toJson(response));
        wsException(response.getMessage(), webSocket);
    }

    @Override
    public void onMessage(SpeechRecognitionResponse response) {
//        System.out.println("onMessage:" + JsonUtil.toJson(response));
    }

    private void wsException(String errorMessage, WebSocket webSocket) {
        WebsocketException websocketException = new WebsocketException(errorMessage);
        webSocket.send(websocketException.toString());
        webSocket.close();
        throw websocketException;
    }

    private String uploadAudio() throws IOException {
        //拼接文件
        List<File> tempSourceFileList = new ArrayList<>();
        for (byte[] value : userAudioByteList) {
            File tempFile = FileUtils.createTempFile(value, ".wav");
            tempSourceFileList.add(tempFile);
        }
        String fileName = UUIDTool.getUUID().toString();
        File tempTargetFile = File.createTempFile(fileName, ".wav");
        WavMergeUtil.mergeWav(tempSourceFileList, tempTargetFile);
        FileInputStream inputStream = new FileInputStream(tempTargetFile);
        byte[] bytes = new byte[inputStream.available()];
        inputStream.read(bytes);

        //上传文件
        String fileUrl = bosConfigService.uploadFileByByteArray2Oss(bytes, fileName, "wav");

        //删除临时文件
        for (File file : tempSourceFileList) {
            file.delete();
        }
        tempTargetFile.delete();
        //关闭流
        inputStream.close();
        return fileUrl;
    }

}
